<?php

interface mtgType
{
  function getName() : string;
  function __toString() : string;
}

class mtgTypeRef implements mtgType
{
  //TODO: not sure if it must be static
  private static $unresolved = array();

  private string $name;
  private ?mtgScope $scope = null;
  private ?mtgType $resolved = null;
  private ?mtgOrigin $origin = null;

  function __construct($name_or_resolved, mtgScope $scope = null, mtgOrigin $origin = null)
  {
    if(is_object($name_or_resolved))
    {
      if(!($name_or_resolved instanceof mtgType))
        throw new Exception("Bad type");
      $this->resolved = $name_or_resolved;
    }
    else
    {
      $this->name = $name_or_resolved;
      if(!$scope)
        throw new Exception("Scope is not set");

      self::$unresolved[] = $this;
    }

    $this->scope = $scope;
    $this->origin = $origin ?? new mtgOrigin();
  }

  function getName() : string
  {
    if($this->resolved)
      return ''.$this->resolved;
    else
      return $this->name;
  }

  static function checkAllResolved()
  {
    while(sizeof(self::$unresolved) > 0)
    {
      $ref = array_shift(self::$unresolved);
      $resolved = $ref->resolve();
      if($resolved instanceof mtgMetaFunc && $resolved->getName())
        throw new Exception("Invalid resolving of named function as type: " . $resolved->getName() . " at " . $ref->origin);
    }
  }

  function resolve() : mtgType
  {
    if($this->resolved)
      return $this->resolved;

    $symb = $this->scope->findSymbol($this->name);
    if($symb)
    {
      if(!($symb instanceof mtgType))
        throw new Exception("{$this->origin} : Symbol '{$this->name}' is not type");
      $this->resolved = $symb;
      return $this->resolved;
    }
    else
      throw new Exception("{$this->origin} : Symbol '{$this->name}' not found");
  } 

  function __toString()
  {
    return $this->getName();
  }
}

abstract class mtgUserType extends mtgMetaUnit implements mtgType
{
  protected string $name;

  function __construct(string $name)
  {
    $this->name = $name;
  }

  function getName() : string
  {
    return $this->name;
  }

  function setName(string $name)
  {
    $this->name = $name;
  }

  function __toString() : string
  {
    return $this->name;
  }
}

class mtgBuiltinType implements mtgType
{
  private string $name;

  function __construct(string $name)
  {
    if(!in_array($name, mtgMetaInfo::$BUILTIN_TYPES))
      throw new Exception("Not a built-in type '$name'");

    $this->name = $name;
  }

  function getName() : string 
  {
    return $this->name;
  }

  function __toString() : string
  {
    return $this->name;
  }

  function isNumeric() : bool
  {
    return !$this->isString() && !$this->isBool() && !$this->isBlob();
  }

  function isString() : bool
  {
    return $this->name === 'string';
  }

  function isInt() : bool
  {
    return strpos($this->name, 'int') === 0;
  }

  function isUint() : bool
  {
    return strpos($this->name, 'uint') === 0;
  }

  function isUint8() : bool
  {
    return $this->name === 'uint8';
  }

  function isUint16() : bool
  {
    return $this->name === 'uint16';
  }

  function isUint32() : bool
  {
    return $this->name === 'uint32';
  }

  function isUint64() : bool
  {
    return $this->name === 'uint64';
  }

  function isInt64() : bool
  {
    return $this->name === 'int64';
  }

  function isInt8() : bool
  {
    return $this->name === 'int8';
  }

  function isInt16() : bool
  {
    return $this->name === 'int16';
  }

  function isInt32() : bool
  {
    return $this->name === 'int32';
  }

  function isFloat() : bool
  {
    return $this->name === 'float';
  }

  function isDouble() : bool
  {
    return $this->name === 'double';
  }

  function isBool() : bool
  {
    return $this->name === 'bool';
  }

  function isBlob() : bool
  {
    return $this->name === 'blob';
  }
}

class mtgMultiType implements mtgType
{
  private $values = array();

  function __construct(array $values = array())
  {
    foreach($values as $v)
      $this->addValue($v);
  }

  function addValue(mtgTypeRef $val)
  {
    $this->values[] = $val;
  }

  function getValues()
  {
    $vals = array();
    foreach($this->values as $val)
      $vals[] = $val->resolve();

    return $vals;
  }

  function getName() : string
  {
    $str = '';
    foreach($this->getValues() as $val)
      $str .= $val . ';';
    return $str;
  }

  function __toString() : string
  {
    return $this->getName();
  }
}

class mtgArrType implements mtgType
{
  private mtgTypeRef $value;

  function __construct(mtgTypeRef $value)
  {
    $this->value = $value;
  }

  function getName() : string
  {
    return $this->getValue() . '[]';
  }

  function __toString() : string
  {
    return $this->getName();
  }

  function getValue() : mixed
  {
    return $this->value->resolve();
  }
}

